#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

void callback(GtkWidget *widget, gpointer callback_data);
void quick_message (gchar *title, gchar *message);
void menu_callback(GtkWidget *widget, gpointer callback_data);
gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data);

/* some global vars, stored addresses that the callback needs to access/change */
GtkTextBuffer *buffer;  	/* the buffer for the text box */
GtkWidget *window;      	/* holds entire interface */
int output = 0;			/* type of output, passed to nbtscan() */
int human = 0;			/* human readable service names, used with verbose */
int quiet = 0;			/* supress banners and error messages in output */
int use137 = 0;			/* use port 137 for traffic, must be root */
char *inputfile = NULL;		/* file to be used for input instead of specified IP range */
char sdelim = '\0';		/* char to be used as script friendly output seperator */
GtkWidget *text0, *text1, *text2, *text3, *text4; /* ip range entry boxes */
GtkWidget *infile_entry;	/* entry used for input file */
GtkWidget *scriptf;		/* stuff for script friendly output */
GtkWidget *scanbutton;		/* scan button, global so we can send focus to it */
GtkWidget *humanbutton;		/* for human readable service names, global so we can show/hide it */
GtkWidget *quietbutton;		/* global so it can be reset */
GtkWidget *p137button;		/* for use137, global so it can be cleared */
GtkWidget *bandwidthtext;	/* for bandwidth amount, global so it can be cleared */
GtkWidget *timeouttext;		/* for timeout value, global so it can be cleared */
GtkWidget *retranstext;		/* for retransmit value, etc */
GtkWidget *selection;		/* file selection box, global so it exists between callback calls */

int main(int argc, char *argv[]) {
	/* general gtk widgets */
	GtkWidget *scrolled;	/* holds the textview box  */
	GtkWidget *table;	/* main layout */
	GtkWidget *box;		/* top left two rows in table */
	GtkWidget *label;	/* "IP Range" and such */
	GtkWidget *button;	/* scan, pause, stop, reset, save, exit */
	GtkWidget *textview;	/* the text box for status and results */
	GtkWidget *vsep;	/* a vertical seperator */
	PangoFontDescription *font_desc; /* for setting the font properties of textview */


	/* widgets for menu */
	GtkWidget *menu;	/* holds menu items */
	GtkWidget *menu_bar;	/* label that pops up a menu */
	GtkWidget *root_menu;	/* the bar itself */
	GtkWidget *menu_items;	/* an item on one of the menus */

	/* call the gtk init function */
	gtk_init(&argc, &argv);
	
	/* create a new window */
	/* ------------------- */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	/* set the window title */
	gtk_window_set_title(GTK_WINDOW (window), "xNBTscan 1.0");
	/* set delete_event handler */
	g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (delete_event), NULL);
	/* set border width */
	gtk_container_set_border_width( GTK_CONTAINER (window), 0);

	/* set the window size */
	gtk_window_set_default_size(GTK_WINDOW (window), 620, 400);

	/* set the window position */
	gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);

	/* create the main layout table, , cells can be different sizes */
	/* -------------------------------------------------------------------------- */
	table = gtk_table_new (5, 1, FALSE);
	/* add the table to the window */
	gtk_container_add (GTK_CONTAINER (window), table);

	/* create the menus */
	/* ---------------- */

	/* the menu bar, holds all the other menu stuff.. */
	menu_bar = gtk_menu_bar_new ();


	/* the file menu label */
	root_menu = gtk_menu_item_new_with_mnemonic ("_File");

	/* file menu */
	menu = gtk_menu_new ();

	menu_items = gtk_menu_item_new_with_mnemonic ("_Save As...");
	/* attach it to the menu */
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
	g_signal_connect(G_OBJECT (menu_items), "activate", G_CALLBACK (menu_callback), "file.save");
	gtk_widget_show (menu_items);

	menu_items = gtk_menu_item_new_with_mnemonic ("E_xit");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
	g_signal_connect(G_OBJECT (menu_items), "activate", G_CALLBACK (delete_event), (gpointer) "exit");
        gtk_widget_show (menu_items);

	/* attach the menu to the it's label */
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (root_menu), menu);

	/* show the label */
	gtk_widget_show (root_menu);
	
	/* attach that label (which includes the menu and items) to the menu bar */
	gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), root_menu);

	/* ------------------------------------------------------- */
	/* the output menu label */
	root_menu = gtk_menu_item_new_with_mnemonic ("_Output");

	/* the output menu */
	menu = gtk_menu_new ();

	menu_items = gtk_menu_item_new_with_mnemonic ("_Normal");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "output.normal");
        gtk_widget_show (menu_items);

	menu_items = gtk_menu_item_new_with_mnemonic ("_Verbose");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "output.verbose");
        gtk_widget_show (menu_items);

	menu_items = gtk_menu_item_new_with_mnemonic ("_Dump Packets");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "output.dump");
        gtk_widget_show (menu_items);

        menu_items = gtk_menu_item_new_with_mnemonic ("/etc/_hosts");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "output.etchosts");
        gtk_widget_show (menu_items);

        menu_items = gtk_menu_item_new_with_mnemonic ("_lmhosts");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "output.lmhosts");
        gtk_widget_show (menu_items);

	menu_items = gtk_menu_item_new_with_mnemonic ("_Script Friendly");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "output.script");
        gtk_widget_show (menu_items);

	/* attach the menu to the right label */
        gtk_menu_item_set_submenu (GTK_MENU_ITEM (root_menu), menu);
        
        /* show the label */
        gtk_widget_show (root_menu);

	/* attach that label (which includes the menu and items) to the menu bar */
        gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), root_menu);


	/* ------------------------------------------------------- */
        /* the Help menu label */
	root_menu = gtk_menu_item_new_with_mnemonic ("_Help");

        /* the output menu */

        menu = gtk_menu_new ();

	menu_items = gtk_menu_item_new_with_mnemonic ("_Contents");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect(G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "help.contents");
        gtk_widget_show (menu_items);

	menu_items = gtk_menu_item_new_with_mnemonic ("_About");
        gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_items);
        g_signal_connect (G_OBJECT (menu_items), "activate",
                                  G_CALLBACK (menu_callback), "help.about");
        gtk_widget_show (menu_items);

	gtk_menu_item_right_justify(GTK_MENU_ITEM (root_menu));

	gtk_menu_item_set_submenu (GTK_MENU_ITEM (root_menu), menu);
	gtk_widget_show (root_menu);
        
        /* attach that label (which includes the menu and items) to the menu bar */
        gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), root_menu);

	/* attach the menu bar to the table */
	gtk_table_attach(GTK_TABLE (table), menu_bar, 0, 1, 0, 1, GTK_FILL | GTK_EXPAND, 0, 0, 0);

	/* show the menu bar */
        gtk_widget_show (menu_bar);


	/* create the horizontal box for the first input row */
	/* ------------------------------------------------- */
	box = gtk_hbox_new(FALSE, 5);

	/* and a label to go in it */
	label = gtk_label_new_with_mnemonic("IP _Range: ");
	gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	/* I tried doing these 5 entries with an array but the signals wouldn't connect */
	/* so here they are the long way */

	/* the first text entry */
	text0 = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL (label), GTK_WIDGET (text0));
	gtk_entry_set_max_length (GTK_ENTRY (text0), 3);
	gtk_entry_set_width_chars(GTK_ENTRY (text0), 3);
	g_signal_connect(G_OBJECT (text0), "changed", G_CALLBACK (callback), (gpointer)"text0");
	gtk_box_pack_start (GTK_BOX (box), text0, FALSE, FALSE, 0);
	gtk_widget_show(text0);

	/* a label . */
	label = gtk_label_new(".");
	gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_CENTER);
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	/* the second text entry */
	text1 = gtk_entry_new();
        gtk_entry_set_max_length (GTK_ENTRY (text1), 3);
        gtk_entry_set_width_chars(GTK_ENTRY (text1), 3);
        g_signal_connect(G_OBJECT (text1), "changed", G_CALLBACK (callback), (gpointer)"text1");
        gtk_box_pack_start (GTK_BOX (box), text1, FALSE, FALSE, 0);
        gtk_widget_show(text1);

	/* a label . */
	label = gtk_label_new(".");
        gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_CENTER);
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	/* the third text entry */
	text2 = gtk_entry_new();
        gtk_entry_set_max_length (GTK_ENTRY (text2), 3);
        gtk_entry_set_width_chars(GTK_ENTRY (text2), 3);
        g_signal_connect(G_OBJECT (text2), "changed", G_CALLBACK (callback), (gpointer)"text2");
        gtk_box_pack_start (GTK_BOX (box), text2, FALSE, FALSE, 0);
        gtk_widget_show(text2);

	/* a label . */
	label = gtk_label_new(".");
        gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_CENTER);
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	/* the fourth  text entry */
	text3 = gtk_entry_new();
        gtk_entry_set_max_length (GTK_ENTRY (text3), 3);
        gtk_entry_set_width_chars(GTK_ENTRY (text3), 3);
        g_signal_connect(G_OBJECT (text3), "changed", G_CALLBACK (callback), (gpointer)"text3");
        gtk_box_pack_start (GTK_BOX (box), text3, FALSE, FALSE, 0);
        gtk_widget_show(text3);

	/* and a label for between the last two text boxes */
        label = gtk_label_new("-");
        gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_CENTER);
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	/* the last IP range entry */
	text4 = gtk_entry_new();
        gtk_entry_set_max_length (GTK_ENTRY (text4), 3);
        gtk_entry_set_width_chars(GTK_ENTRY (text4), 3);
        g_signal_connect(G_OBJECT (text4), "changed", G_CALLBACK (callback), (gpointer)"text4");
        gtk_box_pack_start (GTK_BOX (box), text4, FALSE, FALSE, 0);
        gtk_widget_show(text4);

	/* make a vertical seperator */
	vsep = gtk_vseparator_new();
	gtk_box_pack_start (GTK_BOX (box), vsep, FALSE, FALSE, 10);
	gtk_widget_show(vsep);

	/* a checkbox w/ label for human readable service names */
        humanbutton = gtk_check_button_new_with_mnemonic("H_uman Readable");
        g_signal_connect (G_OBJECT (humanbutton), "toggled", G_CALLBACK (callback), (gpointer) "humanread");
        gtk_box_pack_start (GTK_BOX (box), humanbutton, FALSE, FALSE, 0);
	gtk_widget_set_sensitive(GTK_WIDGET (humanbutton), FALSE);
	gtk_widget_show(humanbutton);

	/* make the Scan button */
        scanbutton = gtk_button_new_with_mnemonic("_Scan");
        g_signal_connect (G_OBJECT (scanbutton), "clicked", G_CALLBACK (callback), (gpointer) "scan");
        gtk_box_pack_end (GTK_BOX (box), scanbutton, FALSE, FALSE, 0);
        gtk_widget_show(scanbutton);

	/* attach that box to the table, top left cell  */
	gtk_table_attach(GTK_TABLE (table), box, 0, 1, 1, 2, GTK_FILL | GTK_EXPAND, 0, 10, 5);
	/* show the box */
        gtk_widget_show(box);

	/* create the horizontal box for the second input row */
	/* -------------------------------------------------- */
        box = gtk_hbox_new(FALSE, 10);

	/* and a label to go in it */
        label = gtk_label_new_with_mnemonic("_Get IPs from file: ");
        gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_LEFT);
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	/* an entry box to type a filename/location */
	infile_entry = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL (label), GTK_WIDGET (infile_entry));
        gtk_entry_set_max_length (GTK_ENTRY (infile_entry), 256);
	gtk_entry_set_width_chars(GTK_ENTRY (infile_entry), 20);
        gtk_box_pack_start (GTK_BOX (box), infile_entry, FALSE, FALSE, 0);
        gtk_widget_show(infile_entry);

	/* a button to offer browsing for the file */
	button = gtk_button_new_with_mnemonic("_...");
        g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (callback), (gpointer) "browse");
        gtk_box_pack_start (GTK_BOX (box), button, FALSE, FALSE, 0);
        gtk_widget_show(button);

	/* a checkbox w/ label for supression of error messages and headers */
        quietbutton = gtk_check_button_new_with_mnemonic("_Quiet");
        g_signal_connect (G_OBJECT (quietbutton), "toggled", G_CALLBACK (callback), (gpointer) "quiet");
        gtk_box_pack_start (GTK_BOX (box), quietbutton, FALSE, FALSE, 0);
        gtk_widget_show(quietbutton);

	/* use port 137, must be root */
	p137button = gtk_check_button_new_with_mnemonic("Us_e 137");
        g_signal_connect (G_OBJECT (p137button), "toggled", G_CALLBACK (callback), (gpointer) "use137");
        gtk_box_pack_start (GTK_BOX (box), p137button, FALSE, FALSE, 0);
	if(geteuid() > 0) 
		gtk_widget_set_sensitive(GTK_WIDGET (p137button), FALSE);
        gtk_widget_show(p137button);

        /* make the Reset button */
        button = gtk_button_new_with_mnemonic("Re_set");
        g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (callback), (gpointer) "reset");
        gtk_box_pack_end (GTK_BOX (box), button, FALSE, FALSE, 0);
        gtk_widget_show(button);

	/* attach that box to the table, second row, left cell */
	gtk_table_attach(GTK_TABLE (table), box, 0, 1, 2, 3, GTK_FILL | GTK_EXPAND, 0, 10, 5);
	/* show the box */
	gtk_widget_show(box);

	/* make the scrolled window */
	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_table_attach(GTK_TABLE(table), scrolled, 0, 1, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 10, 10);
	gtk_widget_show (scrolled);

	/* make the Text View window */
	textview = gtk_text_view_new();
	gtk_text_view_set_editable(GTK_TEXT_VIEW (textview), FALSE);
	font_desc = pango_font_description_from_string ("Fixed 10");
	gtk_widget_modify_font (textview, font_desc);
	pango_font_description_free (font_desc);
	gtk_text_view_set_left_margin(GTK_TEXT_VIEW (textview), 10);
	gtk_text_view_set_right_margin(GTK_TEXT_VIEW (textview), 10);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled), textview);
	gtk_widget_show(textview);

	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));

	/* make the row under the output area, for extra options */
	/* ----------------------------------------------------- */
	box = gtk_hbox_new(FALSE, 5);

	/* make a label and entry box for bandwidth throttling */
        label = gtk_label_new_with_mnemonic("_Bandwidth (bps):");
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        bandwidthtext = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL (label), GTK_WIDGET (bandwidthtext));
        gtk_entry_set_max_length (GTK_ENTRY (bandwidthtext), 13);
        gtk_entry_set_width_chars(GTK_ENTRY (bandwidthtext), 5);
        g_signal_connect(G_OBJECT (bandwidthtext), "changed", G_CALLBACK (callback), (gpointer)"bandwidth");
        gtk_box_pack_start (GTK_BOX (box), bandwidthtext, FALSE, FALSE, 0);
        gtk_widget_show(bandwidthtext);

	/* label and entry for timeout value */
	label = gtk_label_new_with_mnemonic("_Timeout (seconds):");
	gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	timeouttext = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL (label), GTK_WIDGET (timeouttext));
        gtk_entry_set_max_length (GTK_ENTRY (timeouttext), 3);
        gtk_entry_set_width_chars(GTK_ENTRY (timeouttext), 3);
        g_signal_connect(G_OBJECT (timeouttext), "changed", G_CALLBACK (callback), (gpointer)"timeout");
        gtk_box_pack_start (GTK_BOX (box), timeouttext, FALSE, FALSE, 0); 
        gtk_widget_show(timeouttext);

	/* label and entry for retransmit value */
        label = gtk_label_new_with_mnemonic("Retrans_mits:");
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

	retranstext = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL (label), GTK_WIDGET (retranstext));
        gtk_entry_set_max_length (GTK_ENTRY (retranstext), 3);
        gtk_entry_set_width_chars(GTK_ENTRY (retranstext), 3);
	g_signal_connect(G_OBJECT (retranstext), "changed", G_CALLBACK (callback), (gpointer)"retrans");
        gtk_box_pack_start (GTK_BOX (box), retranstext, FALSE, FALSE, 0);
        gtk_widget_show(retranstext);

        /* stuff for script friendly output, don't show this until it's needed */ 
        label = gtk_label_new_with_mnemonic("Se_perator: ");
        gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
	gtk_widget_show(label);
        
        scriptf = gtk_entry_new(); 
	gtk_label_set_mnemonic_widget(GTK_LABEL (label), GTK_WIDGET (scriptf));
        gtk_entry_set_max_length (GTK_ENTRY (scriptf), 1);
        gtk_entry_set_width_chars(GTK_ENTRY (scriptf), 1);
        g_signal_connect(G_OBJECT (scriptf), "changed", G_CALLBACK (callback), (gpointer)"scriptf");
        gtk_box_pack_start (GTK_BOX (box), scriptf, FALSE, FALSE, 0);
	gtk_widget_set_sensitive(GTK_WIDGET (scriptf), FALSE);
	gtk_widget_show(scriptf);

	/* attach that box to the table  */
        gtk_table_attach(GTK_TABLE (table), box, 0, 1, 4, 5, GTK_FILL | GTK_EXPAND, 0, 10, 5);
        /* show the box */
        gtk_widget_show(box);

	/* ----------------------------------------------------- */

	gtk_widget_show (table);
	gtk_widget_show (window);

	gtk_main();

	return (0);
}

void callback(GtkWidget *widget, gpointer data) {
	FILE *nbtscan, *output_file;
	GtkTextIter start, end;
        gchar *contents, *line, *string, *temp;
	int ip_val[5] = {-1,-1,-1,-1,-1}; /* holds contents of text entries */
	int bandwidth = 0;              /* how much bandwidth to use, 0 for default */
	int timeout = 1;                /* timeout to wait for response, default 1 */
	int retrans = 0;                /* retransmits */


	if(strcmp((char *)data, "text0") == 0) {
		string = (gchar *)strdup((char *)gtk_entry_get_text(GTK_ENTRY (text0)));

		if(strlen(string) > 2 && string[strlen(string) - 1] != '.')
			gtk_window_set_focus(GTK_WINDOW (window), text1);
		else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) > 0) {
			string[strlen(string) - 1] = '\0';
			gtk_entry_set_text(GTK_ENTRY (text0), (gchar *)string);
			gtk_window_set_focus(GTK_WINDOW (window), text1);
		} else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) < 1) {
			gtk_entry_set_text(GTK_ENTRY (text0), "");
		}

		free(string);
		string = NULL;
	} else if(strcmp((char *)data, "text1") == 0) {
		string = (gchar *)strdup((char *)gtk_entry_get_text(GTK_ENTRY (text1)));
        
                if(strlen(string) > 2 && string[strlen(string) - 1] != '.')
                        gtk_window_set_focus(GTK_WINDOW (window), text2);
                else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) > 0) {
                        string[strlen(string) - 1] = '\0';
                        gtk_entry_set_text(GTK_ENTRY (text1), (gchar *)string);
                        gtk_window_set_focus(GTK_WINDOW (window), text2);
                } else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) < 1) {
                        gtk_entry_set_text(GTK_ENTRY (text1), "");
                }       
                        
                free(string);
                string = NULL;
	} else if(strcmp((char *)data, "text2") == 0) {
		string = (gchar *)strdup((char *)gtk_entry_get_text(GTK_ENTRY (text2)));
        
                if(strlen(string) > 2 && string[strlen(string) - 1] != '.')
                        gtk_window_set_focus(GTK_WINDOW (window), text3);
                else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) > 0) {
                        string[strlen(string) - 1] = '\0';
                        gtk_entry_set_text(GTK_ENTRY (text2), (gchar *)string);
                        gtk_window_set_focus(GTK_WINDOW (window), text3);
                } else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) < 1) {
                        gtk_entry_set_text(GTK_ENTRY (text2), "");
                }       
                        
                free(string);
                string = NULL;
	} else if(strcmp((char *)data, "text3") == 0) {
		string = (gchar *)strdup((char *)gtk_entry_get_text(GTK_ENTRY (text3)));
        
                if(strlen(string) > 2 && string[strlen(string) - 1] != '.')
                        gtk_window_set_focus(GTK_WINDOW (window), text4);
                else if((string[strlen(string) - 1] == '.' || string[strlen(string) - 1] == '-') && (strlen(string) - 1) > 0) {
                        string[strlen(string) - 1] = '\0';
                        gtk_entry_set_text(GTK_ENTRY (text3), (gchar *)string);
                        gtk_window_set_focus(GTK_WINDOW (window), text4);
                } else if((string[strlen(string) - 1] == '.' || string[strlen(string) - 1] == '-') && (strlen(string) - 1) < 1) {
                        gtk_entry_set_text(GTK_ENTRY (text3), "");
                }       
                        
                free(string);
                string = NULL;
        } else if(strcmp((char *)data, "text4") == 0) {
		string = (gchar *)strdup((char *)gtk_entry_get_text(GTK_ENTRY (text4)));
        
                if(strlen(string) > 2 && string[strlen(string) - 1] != '.')
                        gtk_window_set_focus(GTK_WINDOW (window), scanbutton);
                else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) > 0) {
                        string[strlen(string) - 1] = '\0';
                        gtk_entry_set_text(GTK_ENTRY (text4), (gchar *)string);
                        gtk_window_set_focus(GTK_WINDOW (window), scanbutton);
                } else if(string[strlen(string) - 1] == '.' && (strlen(string) - 1) < 1) {
                        gtk_entry_set_text(GTK_ENTRY (text4), "");
                }       
                        
                free(string);
                string = NULL;
	} else if(strcmp((char *)data, "scan") == 0) {
		sscanf((char *)gtk_entry_get_text(GTK_ENTRY (text0)), "%i", &ip_val[0]);
                sscanf((char *)gtk_entry_get_text(GTK_ENTRY (text1)), "%i", &ip_val[1]);
                sscanf((char *)gtk_entry_get_text(GTK_ENTRY (text2)), "%i", &ip_val[2]);
                sscanf((char *)gtk_entry_get_text(GTK_ENTRY (text3)), "%i", &ip_val[3]);
                sscanf((char *)gtk_entry_get_text(GTK_ENTRY (text4)), "%i", &ip_val[4]);
		sscanf((char *)gtk_entry_get_text(GTK_ENTRY (bandwidthtext)), "%i", &bandwidth);
		sscanf((char *)gtk_entry_get_text(GTK_ENTRY (timeouttext)), "%i", &timeout);
		sscanf((char *)gtk_entry_get_text(GTK_ENTRY (retranstext)), "%i", &retrans);

		if(strlen((gchar *)gtk_entry_get_text(GTK_ENTRY (infile_entry))) > 1) {
			inputfile = (gchar *)strdup((gchar *)gtk_entry_get_text(GTK_ENTRY (infile_entry)));
		}

		if((inputfile == NULL) && (ip_val[0] < 0 || ip_val[1] < 0 || ip_val[2] < 0 || ip_val[3] < 0 || ip_val[4] < 0)) {
			quick_message("Invalid IP Range.", 
			"All five IP values must be completed (with numeric values), or a file must be specified, before initiating a scan.");
			return;
		}

		if(output == 5 && sdelim == '\0') {
			quick_message("Output seperator required.", "Script friendly output mode needs a seperator character.");
			return;
                }

		/* nbtscan [-v] [-d] [-e] [-l] [-t timeout] [-b bandwidth] [-r] 
		[-q] [-s separator] [-m retransmits] (-f filename)|(<scan_range>) */
		/*         1    2    3    4    timeout                     137       
		     5              retrans          input file    range */

		string = (gchar *)g_malloc(sizeof(char)*60);
		strcpy(string, "nbtscan ");

		temp = (gchar *)g_malloc(sizeof(char)*30);

		switch(output) {
		case 1:				/* verbose */
			strcat(string, "-v ");
			if(human == 1)
				strcat(string, "-h ");
			break;
		case 2:				/* dump packets */
			strcat(string, "-d ");
			break;
		case 3:				/* /etc/hosts format */
			strcat(string, "-e ");
			break;
		case 4: 			/* lmhosts format */
			strcat(string, "-l ");
			break;
		case 5: 			/* script friendly */
			strcat(string, "-s ");
			if(strchr("~`#&*()[];\"'\\<>", sdelim))
				strcat(string, "\\\\");
			sprintf(temp, "%c ", (char)sdelim);
			strcat(string, temp);
		}

		if(quiet > 0) 
			strcat(string, "-q ");

		if(use137 > 0)
			strcat(string, "-r "); 

		if(bandwidth > 0) {
			sprintf(temp, "-b %i ", bandwidth);
			strcat(string, temp);
		}

		if(timeout != 1) {
			sprintf(temp, "-t %i ", timeout);
			strcat(string, temp);
                }

		if(retrans > 0) {
                        sprintf(temp, "-m %i ", retrans);
                        strcat(string, temp);
                }

		if(inputfile != NULL) {
			sprintf(temp, "-f %s ", inputfile);
			strcat(string, temp);
		} else {
			sprintf(temp, "%i.%i.%i.%i-%i", ip_val[0], ip_val[1], ip_val[2], ip_val[3], ip_val[4]);
			strcat(string, temp);
		}
		g_free(temp);

		
		nbtscan = popen(string, "r");
		g_free(string);

		if(nbtscan == NULL) {
			quick_message("Failed to execute 'nbtscan'.", 
			"There was an error running nbtscan, please make sure it is installed and the binary is in your path.");
			return;
		}

		line = (gchar *)g_malloc(sizeof(char)*256);
		while(fgets(line, sizeof(line), nbtscan)) 
			gtk_text_buffer_insert_at_cursor(GTK_TEXT_BUFFER (buffer), line, -1);

		/* insert an empty line to seperate output from multiple scans */
		gtk_text_buffer_insert_at_cursor(GTK_TEXT_BUFFER (buffer), "\n\n", -1);

		g_free(line);

		pclose(nbtscan);	
	} else if(strcmp((char *)data, "browse") == 0) {
		selection = gtk_file_selection_new("Get IP range from...");
                g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (selection)->ok_button), "clicked",
                                  G_CALLBACK (callback), "getiprange");
                g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (selection)->cancel_button), "clicked",
                                          G_CALLBACK (gtk_widget_destroy), selection);
                gtk_widget_show(selection);
	} else if(strcmp((char *)data, "getiprange") == 0) {
		// set the file name
		gtk_entry_set_text(GTK_ENTRY (infile_entry), 
				   (gchar *)gtk_file_selection_get_filename(GTK_FILE_SELECTION (selection)));
		gtk_widget_destroy(selection);
        } else if(strcmp((char *)data, "reset") == 0) {
		gtk_entry_set_text(GTK_ENTRY (text0), "");
		gtk_entry_set_text(GTK_ENTRY (text1), "");
		gtk_entry_set_text(GTK_ENTRY (text2), "");
                gtk_entry_set_text(GTK_ENTRY (text3), "");
                gtk_entry_set_text(GTK_ENTRY (text4), "");
		gtk_entry_set_text(GTK_ENTRY (infile_entry), "");
		gtk_entry_set_text(GTK_ENTRY (bandwidthtext), "");
		gtk_entry_set_text(GTK_ENTRY (timeouttext), "");
		gtk_entry_set_text(GTK_ENTRY (retranstext), "");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (quietbutton), FALSE);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (humanbutton), FALSE);
		gtk_text_buffer_set_text(buffer, "", 0);
		if(inputfile != NULL) {
			free(inputfile);
			inputfile = NULL;
		}
	} else if(strcmp((char *)data, "savetofile") == 0) {
		output_file = fopen((gchar *)gtk_file_selection_get_filename(GTK_FILE_SELECTION (selection)), "w");
                if(output_file == NULL) {
                        quick_message("Error Opening File", "Unable to create the specified file.");
                        return;                 
                }       
                        
		gtk_text_buffer_get_bounds(buffer, &start, &end);
                        
                contents = (char *)strdup(gtk_text_buffer_get_text(buffer, &start, &end, FALSE));

                fprintf(output_file, contents);

                fclose(output_file);
                g_free(contents);

		gtk_widget_destroy(selection);
	} else if(strcmp((char *)data, "scriptf") == 0) {
		sscanf((gchar *)gtk_entry_get_text(GTK_ENTRY (scriptf)), "%c", &sdelim);
	} else if(strcmp((char *)data, "humanread") == 0) {
		if(human == 0)
			human = 1;
		else 
			human = 0; 
	} else if(strcmp((char *)data, "quiet") == 0) {
                if(quiet == 0)
			quiet = 1;
		else 
                        quiet = 0;
	} else if(strcmp((char *)data, "use137") == 0) {
                if(use137 == 0)
                        use137 = 1;
                else
                        use137 = 0;
	}
}

void quick_message (gchar *title, gchar *message) {
   GtkWidget *dialog, *label;
   
   /* Create the widgets */
   dialog = gtk_dialog_new_with_buttons (title, GTK_WINDOW (window), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_NONE, NULL);
   label = gtk_label_new (message);
   
   /* Ensure that the dialog box is destroyed when the user responds. */
   g_signal_connect_swapped (GTK_OBJECT (dialog), "response", G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (dialog));

   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);

   /* Add the label, and show everything we've added to the dialog. */
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
   gtk_widget_show_all (dialog);
}

void helpabout (void) {
   GtkWidget *dialog, *label;

   dialog = gtk_dialog_new_with_buttons ("About", GTK_WINDOW (window), GTK_DIALOG_DESTROY_WITH_PARENT, 
		GTK_STOCK_OK, GTK_RESPONSE_NONE, NULL);
   /**************************************************************/
   label = gtk_label_new (
	"NBTscan version 1.0.3. Copyright (C) 1999-2002 Alla Bezroutchko\n"
	"xNBTscan interface version 1.0. Copyright (C) 2002 Bryan Schneiders\n\n"

	"NBTscan is a program for scanning IP networks for NetBIOS name\n"
	"information. It sends NetBIOS status query to each address in\n"
	"supplied range and lists received information in human\n"
	"readable form. For each responded host it lists IP address,\n"
	"NetBIOS computer name, logged-in user name and MAC address\n"
	"(such as Ethernet).\n\n"

	"See http://www.inetcat.org/software/nbtscan.html for\n"
	"NBTscan homepage.\n\n"

	"LICENSE.\n"

	"This program is free software; you can redistribute it and/or modify\n"
	"it under the terms of the GNU General Public License as published by\n"
	"the Free Software Foundation; either version 2 of the License, or\n"
	"(at your option) any later version.\n\n"

	"This program is distributed in the hope that it will be useful,\n"
	"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	"GNU General Public License for more details.\n"

	"You should have received a copy of the GNU General Public License\n"
	"along with this program (in a file called COPYING); if not, write\n"
	"to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,\n"
	"MA 02139, USA.\n");

   /**************************************************************/

   /* Ensure that the dialog box is destroyed when the user responds. */
   g_signal_connect_swapped (GTK_OBJECT (dialog), "response", G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (dialog));

   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);

   /* Add the label, and show everything we've added to the dialog. */
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
   gtk_widget_show_all (dialog);
}

void helpcontents (void) {
        GtkWidget *dialog, *scrolled, *textview;	
	GtkTextBuffer *helpbuf;
	PangoFontDescription *font_desc;

	/* make a dialog "pop-up window" to hold the following.. */
	dialog = gtk_dialog_new_with_buttons ("Help", GTK_WINDOW (window), GTK_DIALOG_DESTROY_WITH_PARENT,
                GTK_STOCK_OK, GTK_RESPONSE_NONE, NULL);

	g_signal_connect_swapped (GTK_OBJECT (dialog), "response", G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (dialog));

	gtk_window_set_default_size(GTK_WINDOW (dialog), 600, 400);

	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);
	
	/* make a scrolled window to hold the help text */
	scrolled = gtk_scrolled_window_new(NULL, NULL);
        gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	/* make the Text View window */
        textview = gtk_text_view_new();
        gtk_text_view_set_editable(GTK_TEXT_VIEW (textview), FALSE);

	/* give it a font */
        font_desc = pango_font_description_from_string ("Fixed 10");
        gtk_widget_modify_font (textview, font_desc);
        pango_font_description_free (font_desc);

	/* set margins */
        gtk_text_view_set_left_margin(GTK_TEXT_VIEW (textview), 10);
        gtk_text_view_set_right_margin(GTK_TEXT_VIEW (textview), 10);

	/* attach it to the scrolled window */
        gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled), textview);
        gtk_widget_show(textview);
   
	/* grab the address of the buffer so we can add to it */ 
        helpbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));

	/* finally add all that to the popup window and show it */
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), scrolled);
	gtk_widget_show (scrolled);
    
	/* adding some help text here would be good, or else we just created a boring window */

	gtk_text_buffer_insert_at_cursor(GTK_TEXT_BUFFER (helpbuf), 
		"Getting started:\n\n"

		"If you just want to go ahead and scan something, enter an ip range such as:\n"
		"192.168.0.1-254 in the \"IP Range\" boxes and click \"Scan\".  You may also use\n"
		"a file containing the target addresses or range by entering it in the\n"
		"\"Get IPs from file:\" box and clicking \"Scan\". If both a file and range are\n"
		"given, the file will be used and the range will be ignored.\n\n"

		"The \"Reset\" button clears all the feilds and resets some values back to their\n"
		"defaults.  The \"Human Readable\" checkbox is only available in \"Verbose\" output\n"
		"mode.  Likewise, the \"Seperator\" feild is only available in \"Script Friendly\"\n"
		"output mode. The \"Use 137\" option is only available if the program is running with\n"
		"super-user priviliges. The options below the output box are optional.\n\n"		

		"Below is the usage for the original nbtscan program, you should be able to\n"
		"figure out how the rest of this works from it.\n\n"

		"Usage:\n\n"

		"nbtscan [-v] [-d] [-e] [-l] [-t timeout] [-b bandwidth] [-r] [-q] [-s separator]\n" 
		"        [-m retransmits] (-f filename)|(<scan_range>)\n"
		"        -v              verbose output. Print all names received\n"
		"                        from each host\n"
		"        -d              dump packets. Print whole packet contents.\n"
		"        -e              Format output in /etc/hosts format.\n"
		"        -l              Format output in lmhosts format.\n"
		"                        Cannot be used with -v, -s or -h options.\n"
		"        -t timeout      wait timeout seconds for response.\n"
		"                        Default 1.\n"
		"        -b bandwidth    Output throttling. Slow down output\n"
		"                        so that it uses no more that bandwidth bps.\n"
		"                        Useful on slow links, so that ougoing queries\n"
		"                        don't get dropped.\n"
		"        -r              use local port 137 for scans. Win95 boxes\n"
		"                        respond to this only.\n"
		"                        You need to be root to use this option on Unix.\n"
		"        -q              Suppress banners and error messages,\n"
		"        -s separator    Script-friendly output. Don't print\n"
		"                        column and record headers, separate fields with\n"
		"                        separator.\n"
		"        -h              Print human-readable names for services.\n"
		"                        Can only be used with -v option.\n"
		"        -m retransmits  Number of retransmits. Default 0.\n"
		"        -f filename     Take IP addresses to scan from file filename\n"
		"        <scan_range>    what to scan. Can either be single IP\n"
		"                        like 192.168.1.1 or\n"
		"                        range of addresses in one of two forms:\n"
		"                        xxx.xxx.xxx.xxx/xx or xxx.xxx.xxx.xxx-xxx.\n\n"
		"Examples:\n"
		"        nbtscan -r 192.168.1.0/24\n"
		"                Scans the whole C-class network.\n"
		"        nbtscan 192.168.1.25-137\n"
		"                Scans a range from 192.168.1.25 to 192.168.1.137\n"
		"        nbtscan -v -s : 192.168.1.0/24\n"
		"                Scans C-class network. Prints results in script-friendly\n"
		"                format using colon as field separator.\n"
		"                Produces output like that:\n"
		"                192.168.0.1:NT_SERVER:00U\n"
		"                192.168.0.1:MY_DOMAIN:00G\n"
		"                192.168.0.1:ADMINISTRATOR:03U\n"
		"                192.168.0.2:OTHER_BOX:00U\n"
                "		 ...\n"
        	"	nbtscan -f iplist\n"
                "		Scans IP addresses specified in file iplist.\n",
		-1);

	gtk_widget_show_all (dialog);
}

void menu_callback(GtkWidget *widget, gpointer data) {
	gchar *lineptr;


	if(strncmp((gchar *)data, "file", 4) == 0) {
		lineptr = data + 5;
		if(strcmp((gchar *)lineptr, "save") == 0) {
			selection = gtk_file_selection_new("Save output to..");
                	g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (selection)->ok_button), "clicked",
                                  G_CALLBACK (callback), "savetofile");
                	g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (selection)->cancel_button), "clicked",
                                  G_CALLBACK (gtk_widget_destroy), selection);
                	gtk_widget_show(selection);
		}
	} else if(strncmp((gchar *)data, "outp", 4) == 0) {
		lineptr = data + 7;

		/* if changing from script to other output
		 * remove the script delim options */
		if(output == 5 && strcmp(lineptr, "script") != 0) {
			gtk_widget_set_sensitive(GTK_WIDGET (scriptf), FALSE);
		}

		if(output == 1 && strcmp(lineptr, "verbose") != 0) {
			gtk_widget_set_sensitive(GTK_WIDGET (humanbutton), FALSE); 
		}

		if(strcmp((gchar *)lineptr, "normal") == 0) {
			output = 0;
                } else if(strcmp((gchar *)lineptr, "verbose") == 0) {
			output = 1;
			gtk_widget_set_sensitive(GTK_WIDGET (humanbutton), TRUE);
                } else if(strcmp((gchar *)lineptr, "dump") == 0) {
			output = 2;
                } else if(strcmp((gchar *)lineptr, "etchosts") == 0) {
			output = 3;
                } else if(strcmp((gchar *)lineptr, "lmhosts") == 0) {
			output = 4;
                } else if(strcmp((gchar *)lineptr, "script") == 0) {
			gtk_widget_set_sensitive(GTK_WIDGET (scriptf), TRUE);
			output = 5;
                }
	} else if(strncmp((gchar *)data, "help", 4) == 0) {
		lineptr = data + 5;

		if(strcmp((gchar *)lineptr, "contents") == 0) {
			helpcontents();
		} else if(strcmp((gchar *)lineptr, "about") == 0) {
			helpabout();
		}
	}
}

gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data) {
	gtk_main_quit();
	return FALSE;
}
